# cython: language_level=3
# cython: profile=True
# Time-stamp: <2024-10-22 15:19:55 Tao Liu>

"""Module for SAPPER ReadAlignment class

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file COPYING included
with the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------
import cython
from cython.cimports.cpython import bool

# cdef extern from "stdlib.h":
#     ctypedef unsigned int size_t
#     size_t strlen(char *s)
#     void *malloc(size_t size)
#     void *calloc(size_t n, size_t size)
#     void free(void *ptr)
#     int strcmp(char *a, char *b)
#     char * strcpy(char *a, char *b)
#     long atol(char *bytes)
#     int atoi(char *bytes)

# ------------------------------------
# constants
# ------------------------------------

__BAMDNACODE__ = b"=ACMGRSVTWYHKDBN"
__CIGARCODE__ = "MIDNSHP=X"
__DNACOMPLEMENT__ = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f !"#$%&\'()*+,-./0123456789:;<=>?@TBGDEFCHIJKLMNOPQRSAUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\x7f\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf\xc0\xc1\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xcb\xcc\xcd\xce\xcf\xd0\xd1\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xdb\xdc\xdd\xde\xdf\xe0\xe1\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xeb\xec\xed\xee\xef\xf0\xf1\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff' # A trans table to convert A to T, C to G, G to C, and T to A.

# -- CIGAR CODE --
#OP BAM  Description
#M  0    alignment match (can be a sequence match or mismatch) insertion to the reference
#I  1    insertion to the reference
#D  2    deletion from the reference
#N  3    skipped region from the reference
#S  4    soft clipping (clipped sequences present in SEQ)
#H  5    hard clipping (clipped sequences NOT present in SEQ)
#P  6    padding (silent deletion from padded reference)
#=  7    sequence match
#X  8    sequence mismatch
# -- -- -- -- -- --

# ------------------------------------
# Misc functions
# ------------------------------------

# ------------------------------------
# Classes
# ------------------------------------


@cython.cclass
class ReadAlignment:
    readname: bytes
    chrom: bytes
    lpos: cython.int
    rpos: cython.int
    # strand information. 0 means forward strand, 1 means reverse strand.
    strand: cython.int
    binaryseq: bytes
    binaryqual: bytes
    l: cython.int                   # length of read
    cigar: tuple             # each item contains op_l|op
    MD: bytes
    # number of edits; higher the number, more differences with reference
    n_edits: cython.int
    SEQ: bytes               # sequence of read regarding to + strand
    QUAL: bytes              # quality of read regarding to + strand

    def __init__(self,
                 readname: bytes,
                 chrom: bytes,
                 lpos: cython.int, rpos: cython.int,
                 strand: cython.int,
                 binaryseq: bytes,
                 binaryqual: bytes,
                 cigar: tuple,
                 MD: bytes):
        self.readname = readname
        self.chrom = chrom
        self.lpos = lpos
        self.rpos = rpos
        self.strand = strand
        self.binaryseq = binaryseq
        self.binaryqual = binaryqual
        self.l = len(binaryqual)
        self.cigar = cigar
        self.MD = MD
        self.n_edits = self.get_n_edits()
        (self.SEQ, self.QUAL) = self.__get_SEQ_QUAL()

    @cython.cfunc
    def get_n_edits(self) -> cython.int:
        """The number is from self.cigar and self.MD.

        """
        n_edits: cython.int
        i: cython.int
        cigar_op: cython.int
        cigar_op_l: cython.int
        c: cython.char

        n_edits = 0
        for i in self.cigar:    # only count insertion or softclip
            cigar_op = i & 15
            cigar_op_l = i >> 4
            if cigar_op in [1, 4]:    # count Insertion or Softclip
                n_edits += cigar_op_l

        for c in self.MD:
            if (c > 64 and c < 91):  # either deletion in query or mismatch
                n_edits += 1
        return n_edits

    def __str__(self):
        c = self.chrom.decode()
        n = self.readname.decode()
        if self.strand:
            s = "-"
        else:
            s = "+"
        return f"{c}\t{self.lpos}\t{self.rpos}\t{n}\t{self.l}\t{s}"
    
    def __getitem__(self, keyname):
        if keyname == "readname":
            return self.readname
        elif keyname == "chrom":
            return self.chrom
        elif keyname == "lpos":
            return self.lpos
        elif keyname == "rpos":
            return self.rpos
        elif keyname == "strand":
            return self.strand
        elif keyname == "SEQ":
            return self.SEQ
        elif keyname == "QUAL":
            return self.QUAL
        elif keyname == "n_edits":
            return self.n_edits
        elif keyname == "binaryseq":
            return self.binaryseq
        elif keyname == "binaryqual":
            return self.binaryqual
        elif keyname == "l":
            return self.l
        elif keyname == "cigar":
            return self.cigar
        elif keyname == "MD":
            return self.MD
        else:
            raise KeyError("No such key", keyname)

    def __getstate__(self):
        return (self.readname, self.chrom, self.lpos, self.rpos, self.strand,
                self.binaryseq, self.binaryqual, self.l, self.cigar,
                self.MD, self.n_edits, self.SEQ, self.QUAL)

    def __setstate__(self, state):
        (self.readname, self.chrom, self.lpos, self.rpos, self.strand,
         self.binaryseq, self.binaryqual, self.l, self.cigar, self.MD,
         self.n_edits, self.SEQ, self.QUAL) = state

    @cython.cfunc
    def __get_SEQ_QUAL(self) -> tuple:
        """Get of: tuple (SEQ, QUAL).

        Rule: for each byte, 1st base in the highest 4bit; 2nd in the
        lowest 4bit. "=ACMGRSVTWYHKDBN" -> [0,15]

        Note: In BAM, if a sequence is mapped to reverse strand, the
        reverse complement seq is written in SEQ field. So the return
        value of this function will not be the original one if the
        read is mapped to - strand. If you need to original one, do
        reversecomp for SEQ and reverse QUAL.

        """
        i: cython.int
        c: cython.char
        seq: bytearray
        qual: bytearray

        seq = bytearray(b"")
        qual = bytearray(b"")

        for i in range(len(self.binaryseq)):
            c = self.binaryseq[i]
            # high
            seq.append(__BAMDNACODE__[c >> 4 & 15])
            # low
            seq.append(__BAMDNACODE__[c & 15])

        for i in range(len(self.binaryqual)):
            # qual is the -10log10 p or phred score.
            qual.append(self.binaryqual[i])

        if seq[-1] == b"=":
            # trim the last '=' if it exists
            seq = seq[:-1]
        assert len(seq) == len(qual), Exception("Lengths of seq and qual are not consistent!")

        # Example on how to get original SEQ and QUAL:
        # if self.strand:
        #    seq.reverse()
        #    #compliment
        #    seq = seq.translate(__DNACOMPLEMENT__)
        #    qual.reverse()

        return (bytes(seq), bytes(qual))

    @cython.ccall
    def get_FASTQ(self) -> bytes:
        """Get FASTQ format text.

        """
        seq: bytes
        qual: bytearray

        seq = self.SEQ
        qual = bytearray(self.QUAL)

        for i in range(len(self.QUAL)):
            # qual is the -10log10 p or phred score, to make FASTQ, we
            # have to add 33
            qual[i] += 33

        # reverse while necessary
        if self.strand:
            seq = self.SEQ[::-1]
            # compliment
            seq = seq.translate(__DNACOMPLEMENT__)
            qual = qual[::-1]
        else:
            seq = self.SEQ

        return b"@" + self.readname + b"\n" + seq + b"\n+\n" + qual + b"\n"

    @cython.ccall
    def get_REFSEQ(self) -> bytearray:
        """Fetch reference sequence, using self.MD and self.cigar
        """
        c: cython.char
        seq: bytearray
        i: cython.int
        cigar_op: cython.int
        cigar_op_l: cython.int
        MD_op: bytearray
        ind: cython.int
        flag_del: bool       # flag for deletion event in query

        # we start with read seq then make modifications
        seq = bytearray(self.SEQ)

        # 2-step proces
        # First step: use CIGAR to edit SEQ to remove S (softclip) and
        # I (insert)
        # __CIGARCODE__ = "MIDNSHP=X"
        # let ind be the index in SEQ
        ind = 0
        for i in self.cigar:
            cigar_op = i & 15
            cigar_op_l = i >> 4
            if cigar_op in [2, 5, 6]:
                # do nothing for Deletion (we will
                # put sequence back in step 2),
                # Hardclip and Padding
                pass
            elif cigar_op in [0, 7, 8]:
                # M = X alignment match (match or mismatch) do nothing
                # and move ind
                ind += cigar_op_l
            elif cigar_op in [1, 4]:    # Remove for Insertion or Softclip
                seq[ind: ind + cigar_op_l] = b''

        # now the seq should be at the same length as rpos-lpos

        # Second step: use MD string to edit SEQ to put back 'deleted
        # seqs' and modify mismatches

        # let ind be the index in SEQ again, from 0
        ind = 0
        MD_op = bytearray(b'')
        flag_del = False
        for c in self.MD:
            if c < 58 and c > 47:
                # means Match
                flag_del = False
                MD_op.append(c)
            elif (c > 64 and c < 91) and not flag_del:
                # An alphabet means Mismatch, Note, if MD is made
                # right, a mismatch should only be 1 letter surrounded
                # by digits.
                ind += int(MD_op)
                seq[ind] = c
                ind += 1
                # reset MD_op
                MD_op = bytearray(b'')
            elif (c > 64 and c < 91) and flag_del:
                seq[ind:ind] = [c,]
                ind += 1
            elif c == 94:
                # means Deletion in query. Now, insert a sequnce into
                # SEQ
                flag_del = True
                ind += int(MD_op)
                # reset MD_op
                MD_op = bytearray(b'')
            else:
                raise Exception("Don't understand this operator in MD: %c" % c)
            # print(seq.decode())

        return seq

    @cython.ccall
    def get_base_by_ref_pos(self, ref_pos: cython.long):
        """Get base by ref position.

        """
        relative_pos: cython.int
        p: cython.int

        assert self.lpos <= ref_pos and self.rpos > ref_pos, Exception("Given position out of alignment location")
        relative_pos = ref_pos - self.lpos
        p = self.relative_ref_pos_to_relative_query_pos(relative_pos)

        if p == -1:             # located in a region deleted in query
            return None
        else:
            return __BAMDNACODE__[(self.binaryseq[p//2] >> ((1-p % 2)*4)) & 15]

    @cython.ccall
    def get_bq_by_ref_pos(self, ref_pos: cython.long):
        """Get base quality by ref position. Base quality is in Phred scale.

        Returned value is the raw Phred-scaled base quality.

        """
        relative_pos: cython.int
        p: cython.int

        assert self.lpos <= ref_pos and self.rpos > ref_pos, Exception("Given position out of alignment location")
        relative_pos = ref_pos - self.lpos
        p = self.relative_ref_pos_to_relative_query_pos(relative_pos)

        if p == -1:             # located in a region deleted in query
            return None
        else:
            return self.binaryqual[p]

    @cython.ccall
    def get_base_bq_by_ref_pos(self, ref_pos: cython.long) -> tuple:
        """Get base and base quality by ref position. Base quality is
        in Phred scale.

        Returned bq is the raw Phred-scaled base quality.

        """
        relative_pos: cython.int
        p: cython.int

        assert self.lpos <= ref_pos and self.rpos > ref_pos, Exception("Given position out of alignment location")
        relative_pos = ref_pos - self.lpos
        p = self.relative_ref_pos_to_relative_query_pos(relative_pos)

        if p == -1:             # located in a region deleted in query
            return None
        else:
            return (__BAMDNACODE__[(self.binaryseq[p//2] >> ((1-p % 2)*4)) & 15],
                    self.binaryqual[p])

    @cython.ccall
    def get_variant_bq_by_ref_pos(self,
                                  ref_pos: cython.long) -> tuple:
        """Get any variants (different with reference) and base
        quality by ref position.

        variants will be

        1) =, if identical

        2) A/T/C/G, if SNV

        3) -, if the reference base is deleted, in this case, bq will
        be the highest possible bq, which is 93.

        4) ^<A/T/C/G>+, if there is an insertion at the location

        Base quality is the raw Phred-scaled base quality.

        """
        i: cython.int
        m: cython.int
        n: cython.int
        res: cython.int
        p: cython.int
        op: cython.int
        op_l: cython.int
        pos: cython.int
        tip: bool
        refseq: bytearray
        p_refseq: bytes
        seq_array: bytearray
        bq_array: bytearray

        assert self.lpos <= ref_pos and self.rpos > ref_pos, Exception("Given position out of alignment location")

        res = ref_pos - self.lpos          # residue
        p = 0
        refseq = self.get_REFSEQ()
        # p_refseq = refseq[res]
        # -- CIGAR CODE --
        # OP BAM  Description
        # M  0    alignment match (can be a sequence match or mismatch) insertion to the reference
        # I  1    insertion to the reference
        # D  2    deletion from the reference
        # N  3    skipped region from the reference
        # S  4    soft clipping (clipped sequences present in SEQ)
        # H  5    hard clipping (clipped sequences NOT present in SEQ)
        # P  6    padding (silent deletion from padded reference)
        # =  7    sequence match
        # X  8    sequence mismatch

        seq_array = bytearray(b'')
        bq_array = bytearray(b'')

        for m in range(len(self.cigar)):
            i = self.cigar[m]
            op = i & 15
            op_l = i >> 4
            if op in [0, 7, 8]:         # M = X alignment match (match or mismatch)
                if res < op_l - 1:
                    # in the range of a CIGAR operator
                    p += res
                    # find the position, now get the ref
                    pos = p
                    seq_array.append(__BAMDNACODE__[(self.binaryseq[p//2] >> ((1-p % 2)*4)) & 15])
                    bq_array.append(self.binaryqual[p])
                    break
                elif res == op_l - 1:
                    p += res
                    pos = p
                    seq_array.append(__BAMDNACODE__[(self.binaryseq[p//2] >> ((1-p % 2)*4)) & 15])
                    bq_array.append(self.binaryqual[p])
                    # now add any insertion later on
                    # get next cigar
                    if m + 1 == len(self.cigar):
                        break
                    i = self.cigar[m + 1]
                    op = i & 15
                    op_l = i >> 4
                    if op == 1:  # insertion
                        for n in range(op_l):
                            p += 1
                            seq_array.append(__BAMDNACODE__[(self.binaryseq[p//2] >> ((1-p % 2)*4)) & 15])
                            bq_array.append(self.binaryqual[p])
                        # prself: cython.int.SEQ, seq_array
                    break
                else:
                    # go to the next cigar code
                    p += op_l
                    res -= op_l
            elif op in [2, 3]:  # D N
                if res < op_l:
                    # find the position, however ...
                    # position located in a region in reference that
                    # not exists in query
                    pos = p
                    seq_array.append(b'*')
                    bq_array.append(93)   # assign 93 for deletion
                    break
                else:
                    # go to the next cigar code
                    res -= op_l
            elif op == 1:      # Insertion
                p += op_l
                # if res == 0:    # no residue left, so return a chunk of inserted sequence
                #     print "shouldn't run this code"
                #     # first, add the insertion point
                #     seq_array = bytearray(b'~')
                #     bq_array.append(self.binaryqual[p])
                #     # then add the inserted seq
                #     for i in range(op_l):
                #         p += 1
                #         seq_array.append(__BAMDNACODE__[(self.binaryseq[p//2] >> ((1-p%2)*4)) & 15] )
                #         bq_array.append(self.binaryqual[p])
                #     break
                # else:
                #     p += op_l
            elif op == 4:      # Softclip. If it's Softclip, we'd better not return the extra seq
                p += op_l

        if pos == 0 or pos == self.l - 1:
            tip = True
        else:
            tip = False
        return (seq_array, bq_array, self.strand, tip, pos)
        # last position ?
        # raise Exception("Not expected to see this")

    @cython.cfunc
    def relative_ref_pos_to_relative_query_pos(self,
                                               relative_ref_pos: cython.long) -> cython.int:
        """Convert relative pos on ref to pos on query.
        """
        p: cython.int
        res: cython.int
        op: cython.int
        op_l: cython.int

        p = 0
        res = relative_ref_pos

        for i in self.cigar:
            op = i & 15
            op_l = i >> 4
            if op in [0, 7, 8]:
                # M = X alignment match (match or mismatch)
                if res < op_l:
                    p += res
                    return p
                else:
                    p += op_l
                    res -= op_l
            elif op in [2, 3]:  # D N
                if res < op_l:
                    # position located in a region in reference that
                    # not exists in query
                    return -1
                else:
                    res -= op_l
            elif op in [1, 4]:       # I
                p += op_l
        return p



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2024-10-22 14:47:22 Tao Liu>

"""Module for SAPPER BAMParser class

Copyright (c) 2017 Tao Liu <tliu4@buffalo.edu>

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file COPYING included
with the distribution).

@status:  experimental
@version: $Revision$
@author:  Tao Liu
@contact: tliu4@buffalo.edu
"""

# ------------------------------------
# python modules
# ------------------------------------
import cython

import cython.cimports.numpy as cnp
from cython.cimports.cpython import bool

from math import log1p, exp, log

LN10 = 2.3025850929940458
LN10_tenth = 0.23025850929940458


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.ccall
def CalModel_Homo(top1_bq_T: cnp.ndarray(cython.int, ndim=1),
                  top1_bq_C: cnp.ndarray(cython.int, ndim=1),
                  top2_bq_T: cnp.ndarray(cython.int, ndim=1),
                  top2_bq_C: cnp.ndarray(cython.int, ndim=1)) -> tuple:
    """Return (lnL, BIC).

    """
    i: cython.int
    lnL: cython.double
    BIC: cython.double

    lnL = 0
    # Phred score is Phred = -10log_{10} E, where E is the error rate.
    # to get the 1-E: 1-E = 1-exp(Phred/-10*M_LN10) = 1-exp(Phred * -LOG10_E_tenth)
    for i in range(top1_bq_T.shape[0]):
        lnL += log1p(-exp(-top1_bq_T[i]*LN10_tenth))
    for i in range(top1_bq_C.shape[0]):
        lnL += log1p(-exp(-top1_bq_C[i]*LN10_tenth))

    for i in range(top2_bq_T.shape[0]):
        lnL += log(exp(-top2_bq_T[i]*LN10_tenth))
    for i in range(top2_bq_C.shape[0]):
        lnL += log(exp(-top2_bq_C[i]*LN10_tenth))

    BIC = -2*lnL                # no free variable, no penalty
    return (lnL, BIC)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.ccall
def CalModel_Heter_noAS(top1_bq_T: cnp.ndarray(cython.int, ndim=1),
                        top1_bq_C: cnp.ndarray(cython.int, ndim=1),
                        top2_bq_T: cnp.ndarray(cython.int, ndim=1),
                        top2_bq_C: cnp.ndarray(cython.int, ndim=1)) -> tuple:
    """Return (lnL, BIC)

    k_T
    k_C
    """
    k_T: cython.int
    k_C: cython.int
    lnL: cython.double
    BIC: cython.double
    tn_T: cython.int
    tn_C: cython.int
    # tn: cython.int  # total observed NTs
    lnL_T: cython.double
    lnL_C: cython.double    # log likelihood for treatment and control

    lnL = 0
    BIC = 0
    # for k_T
    # total oberseved treatment reads from top1 and top2 NTs
    tn_T = top1_bq_T.shape[0] + top2_bq_T.shape[0]

    if tn_T == 0:
        raise Exception("Total number of treatment reads is 0!")
    else:
        (lnL_T, k_T) = GreedyMaxFunctionNoAS(top1_bq_T.shape[0],
                                             top2_bq_T.shape[0],
                                             tn_T,
                                             top1_bq_T,
                                             top2_bq_T)
        lnL += lnL_T
        BIC += -2*lnL_T

    # for k_C
    tn_C = top1_bq_C.shape[0] + top2_bq_C.shape[0]

    if tn_C == 0:
        pass
    else:
        (lnL_C, k_C) = GreedyMaxFunctionNoAS(top1_bq_C.shape[0],
                                             top2_bq_C.shape[0],
                                             tn_C,
                                             top1_bq_C,
                                             top2_bq_C)
        lnL += lnL_C
        BIC += -2*lnL_C

    # tn = tn_C + tn_T

    # we penalize big model depending on the number of reads/samples
    if tn_T == 0:
        BIC += log(tn_C)
    elif tn_C == 0:
        BIC += log(tn_T)
    else:
        BIC += log(tn_T) + log(tn_C)

    return (lnL, BIC)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.ccall
def CalModel_Heter_AS(top1_bq_T: cnp.ndarray(cython.int, ndim=1),
                      top1_bq_C: cnp.ndarray(cython.int, ndim=1),
                      top2_bq_T: cnp.ndarray(cython.int, ndim=1),
                      top2_bq_C: cnp.ndarray(cython.int, ndim=1),
                      max_allowed_ar: cython.float = 0.99) -> tuple:
    """Return (lnL, BIC)

    kc
    ki
    AS_alleleratio
    """
    k_T: cython.int
    k_C: cython.int
    lnL: cython.double
    BIC: cython.double
    tn_T: cython.int
    tn_C: cython.int
    # tn: cython.int  # total observed NTs
    lnL_T: cython.double
    lnL_C: cython.double    # log likelihood for treatment and control
    AS_alleleratio: cython.double   # allele ratio

    lnL = 0
    BIC = 0

    # Treatment
    tn_T = top1_bq_T.shape[0] + top2_bq_T.shape[0]

    if tn_T == 0:
        raise Exception("Total number of treatment reads is 0!")
    else:
        (lnL_T, k_T, AS_alleleratio) = GreedyMaxFunctionAS(top1_bq_T.shape[0],
                                                           top2_bq_T.shape[0],
                                                           tn_T,
                                                           top1_bq_T,
                                                           top2_bq_T,
                                                           max_allowed_ar)
        # print ">>>",lnL_T, k_T, AS_alleleratio
        lnL += lnL_T
        BIC += -2*lnL_T

    # control
    tn_C = top1_bq_C.shape[0] + top2_bq_C.shape[0]

    if tn_C == 0:
        pass
    else:
        # We assume control will not have allele preference
        (lnL_C, k_C) = GreedyMaxFunctionNoAS(top1_bq_C.shape[0],
                                             top2_bq_C.shape[0],
                                             tn_C,
                                             top1_bq_C,
                                             top2_bq_C)
        lnL += lnL_C
        BIC += -2*lnL_C

    # we penalize big model depending on the number of reads/samples
    if tn_T == 0:
        BIC += log(tn_C)
    elif tn_C == 0:
        BIC += 2 * log(tn_T)
    else:
        BIC += 2 * log(tn_T) + log(tn_C)

    return (lnL, BIC)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cfunc
def GreedyMaxFunctionAS(m: cython.int,
                        n: cython.int,
                        tn: cython.int,
                        me: cnp.ndarray(cython.int, ndim=1),
                        ne: cnp.ndarray(cython.int, ndim=1),
                        max_allowed_ar: cython.float = 0.99) -> tuple:
    """Return lnL, k and alleleratio in tuple.

    Note: I only translate Liqing's C++ code into pyx here. Haven't
    done any review.

    """
    dnew: cython.double
    dold: cython.double
    rold: cython.double
    rnew: cython.double
    kold: cython.int
    knew: cython.int
    btemp: bool
    k0: cython.int
    dl: cython.double
    dr: cython.double
    d0: cython.double
    d1l: cython.double
    d1r: cython.double

    assert m+n == tn
    btemp = False
    if tn == 1:                 # only 1 read; I don't expect this to be run...
        dl = calculate_ln(m, n, tn, me, ne, 0, 0)
        dr = calculate_ln(m, n, tn, me, ne, 1, 1)

        if dl > dr:
            k = 0
            return (dl, 0, 0)
        else:
            k = 1
            return (dr, 1, 1)
    elif m == 0:                          # no top1 nt
        return (calculate_ln(m, n, tn, me, ne, 0, m, max_allowed_ar),
                m,
                1-max_allowed_ar)
        # k0 = m + 1
    elif m == tn:                         # all reads are top1
        return (calculate_ln(m, n, tn, me, ne, 1, m, max_allowed_ar),
                m,
                max_allowed_ar)
    else:
        k0 = m

    d0 = calculate_ln(m, n, tn, me, ne, float(k0)/tn, k0, max_allowed_ar)
    d1l = calculate_ln(m, n, tn, me, ne, float(k0-1)/tn, k0-1, max_allowed_ar)
    d1r = calculate_ln(m, n, tn, me, ne, float(k0+1)/tn, k0+1, max_allowed_ar)

    if d0 > d1l-1e-8 and d0 > d1r-1e-8:
        k = k0
        ar = float(k0)/tn
        return (d0, k, ar)
    elif d1l > d0:
        dold = d1l
        kold = k0-1
        rold = float(k0-1)/tn
        while kold > 1:  # disable: when kold=1 still run, than knew=0 is the final run
            knew = kold - 1
            rnew = float(knew)/tn

            dnew = calculate_ln(m,
                                n,
                                tn,
                                me,
                                ne,
                                rnew,
                                knew,
                                max_allowed_ar)

            if (dnew-1e-8 < dold):
                btemp = True
                break
            kold = knew
            dold = dnew
            rold = rnew

        if btemp:               # maximum L value is in [1,m-1];
            k = kold
            ar = rold
            return (dold, k, ar)
        else:                   # L(k=0) is the max for [0,m-1]
            k = kold
            ar = rold
            return (dold, k, ar)

    elif d1r > d0:
        dold = d1r
        kold = k0 + 1
        rold = float(k0 + 1)/tn
        while kold < tn - 1:  # //disable: when kold=tn-1 still run, than knew=tn is the final run
            knew = kold + 1

            rnew = float(knew)/tn

            dnew = calculate_ln(m,
                                n,
                                tn,
                                me,
                                ne,
                                rnew,
                                knew,
                                max_allowed_ar)

            if dnew - 1e-8 < dold:
                btemp = True
                break
            kold = knew
            dold = dnew
            rold = rnew

        if btemp:               # maximum L value is in [m+1,tn-1]
            k = kold
            ar = rold
            return (dold, k, ar)
        else:                   # L(k=tn) is the max for [m+1,tn]
            k = kold
            ar = rold
            return (dold, k, ar)
    else:
        raise Exception("error in GreedyMaxFunctionAS")


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cfunc
def GreedyMaxFunctionNoAS(m: cython.int,
                          n: cython.int,
                          tn: cython.int,
                          me: cnp.ndarray(cython.int, ndim=1),
                          ne: cnp.ndarray(cython.int, ndim=1)) -> tuple:
    """Return lnL, and k in tuple.

    Note: I only translate Liqing's C++ code into pyx here. Haven't
    done any review.

    """
    dnew: cython.double
    dold: cython.double
    kold: cython.int
    knew: cython.int
    btemp: bool
    k0: cython.int
    bg_r: cython.double
    dl: cython.double
    dr: cython.double
    d0: cython.double
    d1l: cython.double
    d1r: cython.double

    btemp = False
    bg_r = 0.5

    if tn == 1:
        dl = calculate_ln(m, n, tn, me, ne, bg_r, 0)
        dr = calculate_ln(m, n, tn, me, ne, bg_r, 1)
        if dl > dr:
            k = 0
            return (dl, 0)
        else:
            k = 1
            return (dr, 1)
    elif m == 0:                          # no top1 nt
        return (calculate_ln(m, n, tn, me, ne, bg_r, m), m)
        # k0 = m + 1
    elif m == tn:                         # all reads are top1
        return (calculate_ln(m, n, tn, me, ne, bg_r, m), m)
    # elif m == 0:
    #    k0 = m + 1
    # elif m == tn:
    #    k0 = m - 1
    else:
        k0 = m

    d0 = calculate_ln(m, n, tn, me, ne, bg_r, k0)
    d1l = calculate_ln(m, n, tn, me, ne, bg_r, k0 - 1)
    d1r = calculate_ln(m, n, tn, me, ne, bg_r, k0 + 1)

    if d0 > d1l - 1e-8 and d0 > d1r - 1e-8:
        k = k0
        return (d0, k)
    elif d1l > d0:
        dold = d1l
        kold = k0 - 1
        while kold >= 1:   # //when kold=1 still run, than knew=0 is the final run
            knew = kold - 1
            dnew = calculate_ln(m, n, tn, me, ne, bg_r, knew)
            if dnew - 1e-8 < dold:
                btemp = True
                break
            kold = knew
            dold = dnew
                        
        if btemp:               # //maximum L value is in [1,m-1];
            k = kold
            return (dold, k)
        else:                   # //L(k=0) is the max for [0,m-1]
            k = kold
            return (dold, k)
    elif d1r > d0:
        dold = d1r
        kold = k0 + 1
        while kold <= tn - 1:  # //when kold=tn-1 still run, than knew=tn is the final run
            knew = kold + 1
            dnew = calculate_ln(m, n, tn, me, ne, bg_r, knew)
            if dnew - 1e-8 < dold:
                btemp = True
                break
            kold = knew
            dold = dnew

        if btemp:               # //maximum L value is in [m+1,tn-1]
            k = kold
            return (dold, k)
        else:                   # //L(k=tn) is the max for [m+1,tn]
            k = kold
            return (dold, k)
    else:
        raise Exception("error in GreedyMaxFunctionNoAS")


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cfunc
def calculate_ln(m: cython.int,
                 n: cython.int,
                 tn: cython.int,
                 me: cnp.ndarray(cython.int, ndim=1),
                 ne: cnp.ndarray(cython.int, ndim=1),
                 r: cython.double,
                 k: cython.int,
                 max_allowed_r: cython.float = 0.99):
    """Calculate log likelihood given quality of top1 and top2, the
    ratio r and the observed k.

    """
    i: cython.int
    lnL: cython.double
    e: cython.double

    lnL = 0

    # r is extremely high or
    if r > max_allowed_r or r < 1 - max_allowed_r:
        lnL += k*log(max_allowed_r) + (tn-k)*log(1 - max_allowed_r)
    else:
        lnL += k*log(r) + (tn-k)*log(1-r)

    # it's entirely biased toward 1 allele
    if k == 0 or k == tn:
        pass
    elif k <= tn/2:
        for i in range(k):
            lnL += log(float(tn-i)/(k-i))
    else:
        for i in range(tn-k):
            lnL += log(float(tn-i)/(tn-k-i))

    for i in range(m):
        e = exp(- me[i] * LN10_tenth)
        lnL += log((1-e)*(float(k)/tn) + e*(1-float(k)/tn))

    for i in range(n):
        e = exp(- ne[i] * LN10_tenth)
        lnL += log((1-e)*(1-float(k)/tn) + e*(float(k)/tn))

    return lnL


@cython.ccall
def calculate_GQ(lnL1: cython.double,
                 lnL2: cython.double,
                 lnL3: cython.double) -> cython.int:
    """GQ1 = -10*log_{10}((L2+L3)/(L1+L2+L3))
    """
    L1: cython.double
    L2: cython.double
    L3: cython.double
    s: cython.double
    tmp: cython.double
    GQ_score: cython.int

    # L1 = exp(lnL1-lnL1)
    L1 = 1
    L2 = exp(lnL2-lnL1)
    L3 = exp(lnL3-lnL1)

    # if L1 > 1:
    #    L1 = 1

    if L2 > 1:
        L2 = 1
    if L3 > 1:
        L3 = 1
    # if(L1<1e-110) L1=1e-110;
    if L2 < 1e-110:
        L2 = 1e-110
    if L3 < 1e-110:
        L3 = 1e-110

    s = L1 + L2 + L3
    tmp = (L2 + L3)/s
    if tmp > 1e-110:
        GQ_score = (int)(-4.34294*log(tmp))
    else:
        GQ_score = 255

    return GQ_score


@cython.ccall
def calculate_GQ_heterASsig(lnL1: cython.double,
                            lnL2: cython.double) -> cython.int:
    """
    """
    L1: cython.double
    L2: cython.double
    s: cython.double
    tmp: cython.double
    ASsig_score: cython.int

    # L1=exp(2.7182818,lnL1-lnL1)
    L1 = 1
    L2 = exp(lnL2 - lnL1)

    # if L1 > 1:
    #    L1 = 1
    if L2 > 1:
        L2 = 1
    # if L1 < 1e-110:
    #    L1 = 1e-110
    if L2 < 1e-110:
        L2 = 1e-110

    s = L1 + L2
    tmp = L2/s
    if tmp > 1e-110:
        ASsig_score = (int)(-4.34294*log(tmp))
    else:
        ASsig_score = 255

    return ASsig_score



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2024-10-22 16:26:57 Tao Liu>

"""Module for ReadAlignment collection

Copyright (c) 2024 Tao Liu

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file COPYING included
with the distribution).

@status:  experimental
@version: $Revision$
@author:  Tao Liu
@contact: vladimir.liu@gmail.com
"""
# ------------------------------------
# python modules
# ------------------------------------
from collections import Counter
from operator import itemgetter
from copy import copy

from MACS3.Signal.ReadAlignment import ReadAlignment
from MACS3.Signal.PosReadsInfo import PosReadsInfo
from MACS3.Signal.UnitigRACollection import UnitigRAs, UnitigCollection
from MACS3.IO.PeakIO import PeakIO

import cython
from cython.cimports.cpython import bool
# from cython.cimports.cpython.mem import PyMem_Malloc, PyMem_Free

from cython.cimports.libc.stdlib import malloc, free

# ------------------------------------
# constants
# ------------------------------------
__version__ = "Parser $Revision$"
__author__ = "Tao Liu <tliu4@buffalo.edu>"
__doc__ = "All Parser classes"

__DNACOMPLEMENT__ = b'\x00\x01\x02\x03\x04\x05\x06\x07\x08\t\n\x0b\x0c\r\x0e\x0f\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f !"#$%&\'()*+,-./0123456789:;<=>?@TBGDEFCHIJKLMNOPQRSAUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~\x7f\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf\xc0\xc1\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xcb\xcc\xcd\xce\xcf\xd0\xd1\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xdb\xdc\xdd\xde\xdf\xe0\xe1\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xeb\xec\xed\xee\xef\xf0\xf1\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff' # A trans table to convert A to T, C to G, G to C, and T to A.

__CIGARCODE__ = "MIDNSHP=X"

# ------------------------------------
# Misc functions
# ------------------------------------

# ------------------------------------
# Classes
# ------------------------------------


@cython.cclass
class RACollection:
    """A collection of ReadAlignment objects and the corresponding
    PeakIO.

    """
    chrom: bytes
    peak: PeakIO             # A PeakIO object
    RAlists: list           # contain ReadAlignment lists for treatment (0) and control (1)
    left: cython.long               # left position of peak
    right: cython.long              # right position of peak
    length: cython.long             # length of peak
    RAs_left: cython.long           # left position of all RAs in the collection
    RAs_right: cython.long          # right position of all RAs in the collection
    sorted: bool             # if sorted by lpos
    peak_refseq: bytes       # reference sequence in peak region b/w left and right
    peak_refseq_ext: bytes   # reference sequence in peak region with extension on both sides b/w RAs_left and RAs_right

    def __init__(self, chrom: bytes, peak: PeakIO, RAlist_T: list, RAlist_C: list = []):
        """Create RACollection by: object taking:

        1. peak: a PeakIO indicating: object the peak region.

        2. RAlist: a python of: list ReadAlignment objects containing
        all the reads overlapping the peak region. If no RAlist_C
        given, it will be [].

        """
        if len(RAlist_T) == 0:
            # no reads, return None
            raise Exception("No reads from ChIP sample to construct RAcollection!")
        self.chrom = chrom
        self.peak = peak
        # print(len(RAlist_T),"\n")
        # print(len(RAlist_C),"\n")
        self.RAlists = [RAlist_T, RAlist_C]
        self.left = peak["start"]
        self.right = peak["end"]
        self.length = self.right - self.left
        if RAlist_T:
            self.RAs_left = RAlist_T[0]["lpos"]  # initial assignment of RAs_left
            self.RAs_right = RAlist_T[-1]["rpos"]  # initial assignment of RAs_right
            self.sort()                           # it will set self.sorted = True
        else:
            self.RAs_left = -1
            self.RAs_right = -1
        # check RAs_left and RAs_right
        for ra in RAlist_T:
            if ra["lpos"] < self.RAs_left:
                self.RAs_left = ra["lpos"]
            if ra["rpos"] > self.RAs_right:
                self.RAs_right = ra["rpos"]

        for ra in RAlist_C:
            if ra["lpos"] < self.RAs_left:
                self.RAs_left = ra["lpos"]
            if ra["rpos"] > self.RAs_right:
                self.RAs_right = ra["rpos"]
        (self.peak_refseq, self.peak_refseq_ext) = self.__get_peak_REFSEQ()

    def __getitem__(self, keyname):
        if keyname == "chrom":
            return self.chrom
        elif keyname == "left":
            return self.left
        elif keyname == "right":
            return self.right
        elif keyname == "RAs_left":
            return self.RAs_left
        elif keyname == "RAs_right":
            return self.RAs_right
        elif keyname == "length":
            return self.length
        elif keyname == "count":
            return len(self.RAlists[0]) + len(self.RAlists[1])
        elif keyname == "count_T":
            return len(self.RAlists[0])
        elif keyname == "count_C":
            return len(self.RAlists[1])
        elif keyname == "peak_refseq":
            return self.peak_refseq
        elif keyname == "peak_refseq_ext":
            return self.peak_refseq_ext
        else:
            raise KeyError("Unavailable key:", keyname)

    def __getstate__(self):
        #return {"chrom":self.chrom, "peak":self.peak, "RAlists":self.RAlists,
        #        "left":self.left, "right":self.right, "length": self.length,
        #        "RAs_left":self.RAs_left, "RAs_right":self.RAs_right}
        return (self.chrom, self.peak, self.RAlists, self.left, self.right,
                self.length, self.RAs_left, self.RAs_right, self.peak_refseq,
                self.peak_refseq_ext)

    def __setstate__(self, state):
        (self.chrom, self.peak, self.RAlists, self.left, self.right,
         self.length, self.RAs_left, self.RAs_right, self.peak_refseq,
         self.peak_refseq_ext) = state

    @cython.ccall
    def sort(self):
        """Sort RAs according to lpos. Should be used after realignment.

        """
        if self.RAlists[0]:
            self.RAlists[0].sort(key=itemgetter("lpos"))
        if self.RAlists[1]:
            self.RAlists[1].sort(key=itemgetter("lpos"))
        self.sorted = True
        return

    @cython.ccall
    def remove_outliers(self, percent: cython.int = 5):
        """ Remove outliers with too many n_edits. The outliers with
        n_edits in top p% will be removed.

        Default: remove top 5% of reads that have too many differences
        with reference genome.
        """
        n_edits_list: list
        ralist: list
        read: ReadAlignment         # ReadAlignment object
        highest_n_edits: cython.int
        new_RAlist: list
        i: cython.int

        n_edits_list = []
        for ralist in self.RAlists:
            for read in ralist:
                n_edits_list.append(read["n_edits"])
        n_edits_list.sort()
        highest_n_edits = n_edits_list[int(len(n_edits_list) * (1 - percent * .01))]

        for i in (range(len(self.RAlists))):
            new_RAlist = []
            for read in self.RAlists[i]:
                if read["n_edits"] <= highest_n_edits:
                    new_RAlist.append(read)
            self.RAlists[i] = new_RAlist

        return

    @cython.ccall
    def n_edits_sum(self) -> cython.int:
        """
        """
        n_edits_list: list
        ralist: list
        read: ReadAlignment
        c: cython.int
        # highest_n_edits: cython.int

        n_edits_list = []

        for ralist in self.RAlists:
            for read in ralist:
                n_edits_list.append(read["n_edits"])

        n_edits_list.sort()
        # print (n_edits_list)
        c = Counter(n_edits_list)
        return c
        # print(c)

    @cython.cfunc
    def __get_peak_REFSEQ(self) -> tuple:
        """Get the reference sequence within the peak region.

        """
        peak_refseq: bytearray
        # i: cython.int
        # prev_r: cython.long                   #remember the previous filled right end
        start: cython.long
        end: cython.long
        # ind: cython.long
        # ind_r: cython.long
        # read: ReadAlignment
        # read_refseq_ext: bytearray
        # read_refseq: bytearray

        start = min(self.RAs_left, self.left)
        end = max(self.RAs_right, self.right)
        # print ("left",start,"right",end)
        peak_refseq_ext = bytearray(b'N' * (end - start))

        # for treatment.
        peak_refseq_ext = self.__fill_refseq(peak_refseq_ext,
                                             self.RAlists[0])
        # and control if available.
        if self.RAlists[1]:
            peak_refseq_ext = self.__fill_refseq(peak_refseq_ext,
                                                 self.RAlists[1])

        # trim
        peak_refseq = peak_refseq_ext[self.left - start: self.right - start]
        return (bytes(peak_refseq), bytes(peak_refseq_ext))

    @cython.cfunc
    def __fill_refseq(self,
                      seq: bytearray,
                      ralist: list) -> bytearray:
        """Fill refseq sequence of whole peak with refseq sequences of
        each read in ralist.

        """
        prev_r: cython.long         # previous right position of last
        # filled
        ind: cython.long
        ind_r: cython.long
        start: cython.long
        # end: cython.long
        read: ReadAlignment
        read_refseq: bytearray

        start = min(self.RAs_left, self.left)

        # print(len(ralist),"\n")
        prev_r = ralist[0]["lpos"]

        for i in range(len(ralist)):
            read = ralist[i]
            if read["lpos"] > prev_r:
                read = ralist[i - 1]
                read_refseq = read.get_REFSEQ()
                ind = read["lpos"] - start
                ind_r = ind + read["rpos"] - read["lpos"]
                seq[ind: ind_r] = read_refseq
                prev_r = read["rpos"]
        # last
        read = ralist[-1]
        read_refseq = read.get_REFSEQ()
        ind = read["lpos"] - start
        ind_r = ind + read["rpos"] - read["lpos"]
        seq[ind: ind_r] = read_refseq
        return seq

    @cython.ccall
    def get_PosReadsInfo_ref_pos(self,
                                 ref_pos: cython.long,
                                 ref_nt: bytes,
                                 Q: cython.int = 20):
        """Generate a PosReadsInfo for: object a given reference genome
        position.

        Return a PosReadsInfo object.

        """
        s: bytearray
        bq: bytearray
        strand: cython.int
        ra: ReadAlignment
        # bq_list_t: list = []
        # bq_list_c: list = []
        i: cython.int
        pos: cython.int
        tip: bool
        posreadsinfo_p: PosReadsInfo

        posreadsinfo_p = PosReadsInfo(ref_pos, ref_nt)

        # Treatment group
        for i in range(len(self.RAlists[0])):
            ra = self.RAlists[0][i]
            if ra["lpos"] <= ref_pos and ra["rpos"] > ref_pos:
                (s, bq, strand, tip, pos) = ra.get_variant_bq_by_ref_pos(ref_pos)
                posreadsinfo_p.add_T(i, bytes(s), bq[0], strand, tip, Q=Q)

        # Control group
        for i in range(len(self.RAlists[1])):
            ra = self.RAlists[1][i]
            if ra["lpos"] <= ref_pos and ra["rpos"] > ref_pos:
                (s, bq, strand, tip, pos) = ra.get_variant_bq_by_ref_pos(ref_pos)
                posreadsinfo_p.add_C(i, bytes(s), bq[0], strand, Q=Q)

        return posreadsinfo_p

    @cython.ccall
    def get_FASTQ(self) -> bytearray:
        """Get FASTQ file for all reads in RACollection.

        """
        ra: ReadAlignment
        fastq_text: bytearray

        fastq_text = bytearray(b"")

        for ra in self.RAlists[0]:
            fastq_text += ra.get_FASTQ()

        for ra in self.RAlists[1]:
            fastq_text += ra.get_FASTQ()

        return fastq_text

    @cython.cfunc
    def fermi_assemble(self,
                       fermiMinOverlap: cython.int,
                       opt_flag: cython.int = 0x80) -> list:
        """A wrapper function to call Fermi unitig building functions.
        """
        opt: cython.pointer(fml_opt_t)
        # c: cython.int
        n_seqs: cython.int
        n_utg: cython.pointer(cython.int)
        seqs: cython.pointer(bseq1_t)
        utg: cython.pointer(fml_utg_t)
        p: fml_utg_t

        # unitig_k: cython.int
        # merge_min_len: cython.int
        tmps: bytes
        tmpq: bytes
        # ec_k: cython.int = -1
        l: cython.long
        cseq: cython.pointer(cython.char)
        cqual: cython.pointer(cython.char)
        i: cython.int
        j: cython.int
        # tmpunitig: bytes
        # unitig: bytes                 # final unitig
        unitig_list: list             # contain of: list sequences in format: bytes
        # n: cython.pointer(cython.int)

        n_seqs = len(self.RAlists[0]) + len(self.RAlists[1])

        # prn_seqs: cython.int

        # prepare seq and qual, note, we only extract SEQ according to the +
        # strand of reference sequence.
        seqs = cython.cast(cython.pointer(bseq1_t),
                           malloc(n_seqs * cython.sizeof(bseq1_t)))  # we rely on fermi-lite to free this mem

        i = 0
        for ra in self.RAlists[0]:
            tmps = ra["SEQ"]
            tmpq = ra["QUAL"]
            l = len(tmps)
            # we rely on fermi-lite to free this mem
            cseq = cython.cast(cython.pointer(cython.char),
                               malloc((l+1)*cython.sizeof(cython.char)))
            # we rely on fermi-lite to free this mem
            cqual = cython.cast(cython.pointer(cython.char),
                                malloc((l+1)*cython.sizeof(cython.char)))
            for j in range(l):
                cseq[j] = tmps[j]
                cqual[j] = tmpq[j] + 33
            cseq[l] = b'\x00'
            cqual[l] = b'\x00'

            seqs[i].seq = cseq
            seqs[i].qual = cqual
            seqs[i].l_seq = len(tmps)
            i += 1

            # print "@",ra["readname"].decode()
            # prcseq: cython.int.decode()
            # print "+"
            # prcqual: cython.int.decode()

        for ra in self.RAlists[1]:
            tmps = ra["SEQ"]
            tmpq = ra["QUAL"]
            l = len(tmps)
            # we rely on fermi-lite to free this mem
            cseq = cython.cast(cython.pointer(cython.char),
                               malloc((l+1)*cython.sizeof(cython.char)))
            # we rely on fermi-lite to free this mem
            cqual = cython.cast(cython.pointer(cython.char),
                                malloc((l+1)*cython.sizeof(cython.char)))
            for j in range(l):
                cseq[j] = tmps[j]
                cqual[j] = tmpq[j] + 33
            cseq[l] = b'\x00'
            cqual[l] = b'\x00'

            seqs[i].seq = cseq
            seqs[i].qual = cqual
            seqs[i].l_seq = len(tmps)
            i += 1
            # print "@",ra["readname"].decode()
            # prcseq: cython.int.decode()
            # print "+"
            # prcqual: cython.int.decode()

        # if self.RAlists[1]:
        #     unitig_k=int(min(self.RAlists[0][0]["l"],self.RAlists[1][0]["l"])*fermiOverlapMinRatio)

        #     merge_min_len=int(min(self.RAlists[0][0]["l"],self.RAlists[1][0]["l"])*0.5)
        # else:
        #     unitig_k = int(self.RAlists[0][0]["l"]*fermiOverlapMinRatio)

        #     merge_min_len=int(self.RAlists[0][0]["l"]*0.5)
        #fermiMinOverlap = int(self.RAlists[0][0]["l"]*fermiOverlapMinRatio)

        # minimum overlap to merge, default 0
        # merge_min_len= max(25, int(self.RAlists[0][0]["l"]*0.5))
        # merge_min_len= int(self.RAlists[0][0]["l"]*0.5)

        # opt = cython.cast(cython.pointer(fml_opt_t),
        #                   PyMem_Malloc(cython.sizeof(fml_opt_t)))
        # n_utg = cython.cast(cython.pointer(cython.int),
        #                     PyMem_Malloc(cython.sizeof(int)))

        opt = cython.cast(cython.pointer(fml_opt_t),
                          malloc(cython.sizeof(fml_opt_t)))
        n_utg = cython.cast(cython.pointer(cython.int),
                            malloc(cython.sizeof(int)))

        fml_opt_init(opt)
        # k-mer length for error correction (0 for auto; -1 to disable)
        # opt.ec_k = 0

        # min overlap length during initial assembly
        opt.min_asm_ovlp = fermiMinOverlap

        # minimum length to merge, during assembly, don't explicitly merge an overlap if shorter than this value
        # opt.min_merge_len = merge_min_len

        # there are more 'options' for mag clean:
        # flag, min_ovlp, min_elen, min_ensr, min_insr, max_bdist, max_bdiff, max_bvtx, min_merge_len, trim_len, trim_depth, min_dratio1, max_bcov, max_bfrac
        # min_elen (300) will be adjusted
        # min_ensr (4), min_insr (3) will be computed
        # min_merge_len (0) will be updated using opt.min_merge_len

        # We can adjust: flag (0x40|0x80), min_ovlp (0), min_dratio1 (0.7), max_bdiff (50), max_bdist (512), max_bvtx (64), trim_len (0), trim_depth (6), max_bcov (10.), max_bfrac (0.15)

        # 0x20: MAG_F_AGGRESSIVE pop variant bubbles
        # 0x40: MAG_F_POPOPEN aggressive tip trimming
        # 0x80: MAG_F_NO_SIMPL skip bubble simplification
        opt.mag_opt.flag = opt_flag

        # mag_opt.min_ovlp
        #opt.mag_opt.min_ovlp = fermiMinOverlap

        # drop an overlap if its length is below maxOvlpLen*FLOAT
        #opt.mag_opt.min_dratio1 = 0.5

        # retain a bubble if one side is longer than the other side by >INT-bp
        #opt.mag_opt.max_bdiff = 10#merge_min_len

        # trim_len:
        # trim_depth: Parameter used to trim the open end/tip. If trim_len == 0, do nothing

        # max_bdist:
        # max_bvtx: Parameter used to simply bubble while 0x80 flag is set.
        #opt.mag_opt.max_bdist = 1024
        #opt.mag_opt.max_bvtx = 128

        # max_bcov:
        # max_bfrac: Parameter used when aggressive bubble removal is not used. Bubble will be removed if its average coverage lower than max_bcov and fraction (cov1/(cov1+cov2)) is lower than max_bfrac
        #opt.mag_opt.max_bcov = 10.
        #opt.mag_opt.max_bfrac = 0.01

        utg = fml_assemble(opt, n_seqs, seqs, n_utg)
        # get results
        unitig_list = []
        for i in range(n_utg[0]):
            p = utg[i]
            if (p.len < 0):
                continue
            # unitig = b''
            # for j in range(p.len):
            #    unitig += [b'A',b'C',b'G',b'T',b'N'][int(p.seq[j]) - 1]
            # unitig_list.append(unitig)
            unitig_list.append(p.seq)

        fml_utg_destroy(n_utg[0], utg)

        # PyMem_Free(opt)
        # PyMem_Free(n_utg)
        free(opt)
        free(n_utg)

        return unitig_list

    @cython.cfunc
    def align_unitig_to_REFSEQ(self, unitig_list: list) -> tuple:
        """Note: we use smith waterman, but we don't use linear gap
        penalty at this time.

        Also, if unitig is mapped to - strand, we will revcomp the
        unitig. So the unitig_will: list be changed in this case.
        """
        unitig: bytes
        problem: seq_pair_t
        results: cython.pointer(align_t)
        # tmp: cython.pointer(cython.char)
        target: bytes
        reference: bytes
        target_aln_f: bytes
        target_aln_r: bytes
        reference_aln_f: bytes
        reference_aln_r: bytes
        markup_aln_f: bytes
        markup_aln_r: bytes
        score_f: cython.double
        score_r: cython.double
        target_alns: list = []
        reference_alns: list = []
        markup_alns: list = []
        aln_scores: list = []
        i: cython.int

        reference = copy(self.peak_refseq_ext+b'\x00')

        for i in range(len(unitig_list)):
            unitig = unitig_list[i]
            target = copy(unitig + b'\x00')
            # we use swalign.c for local alignment (without affine gap
            # penalty). Will revise later.
            problem.a = target
            problem.alen = len(unitig)
            problem.b = reference
            problem.blen = len(self.peak_refseq_ext)
            results = smith_waterman(cython.address(problem))
            target_aln_f = results.seqs.a
            reference_aln_f = results.seqs.b
            markup_aln_f = results.markup
            score_f = results.score
            free(results.seqs.a)
            free(results.seqs.b)
            free(results.markup)
            free(results)
            # end of local alignment

            # try reverse complement
            target = copy(unitig[::-1] + b'\x00')
            target = target.translate(__DNACOMPLEMENT__)
            problem.a = target
            problem.alen = len(unitig)
            problem.b = reference
            problem.blen = len(self.peak_refseq_ext)
            results = smith_waterman(cython.address(problem))
            target_aln_r = results.seqs.a
            reference_aln_r = results.seqs.b
            markup_aln_r = results.markup
            score_r = results.score
            free(results.seqs.a)
            free(results.seqs.b)
            free(results.markup)
            free(results)
            # end of local alignment

            if score_f > score_r:
                target_alns.append(target_aln_f)
                reference_alns.append(reference_aln_f)
                markup_alns.append(markup_aln_f)
                aln_scores.append(score_f)
            else:
                target_alns.append(target_aln_r)
                reference_alns.append(reference_aln_r)
                markup_alns.append(markup_aln_r)
                aln_scores.append(score_r)
                # we will revcomp unitig
                unitig = unitig[::-1]
                unitig_list[i] = unitig.translate(__DNACOMPLEMENT__)

        return (target_alns, reference_alns, aln_scores, markup_alns)

    @cython.cfunc
    def verify_alns(self, unitig_list, unitig_alns, reference_alns, aln_scores, markup_alns, min_score_100: cython.float = 150):
        """Remove aln/unitig if it contains too many edits in a small region

        default min score is 150, which means under 2/-3/-5/-2 scoring schema, there are 10 mismatches within 100bps region.
        """
        i: cython.int

        for i in range(len(unitig_list)-1, -1, -1):
            # pri: cython.int, aln_scores[i]
            # prunitig_alns: cython.int[i]
            # prmarkup_alns: cython.int[i]
            # prreference_alns: cython.int[i]
            if aln_scores[i] * 100 / len(markup_alns[i]) < min_score_100:
                unitig_list.pop(i)
                unitig_alns.pop(i)
                reference_alns.pop(i)
                aln_scores.pop(i)
                markup_alns.pop(i)
        return

    @cython.cfunc
    def filter_unitig_with_bad_aln(self, unitig_list: list,
                                   target_alns: list,
                                   reference_alns: list,
                                   gratio: float = 0.25) -> tuple:
        """Remove unitigs that has too much gaps (both on target and
        reference) during alignments.

        """
        pass

    @cython.cfunc
    def remap_RAs_w_unitigs(self, unitig_list: list) -> list:
        """Remap RAs to unitigs, requiring perfect match.

        Return RAlists_T, RAlists_C, unmapped_racollection.
        """
        RAlists_T: list = []  # lists of of: list RAs of ChIP mapped to each unitig
        RAlists_C: list = []
        unmapped_RAlist_T: list = []  # of: list RAs of ChIP unmappable to unitigs
        unmapped_RAlist_C: list = []
        # RACollection unmapped_ra_collection
        flag: cython.int = 0
        i: cython.int
        tmp_ra: ReadAlignment
        tmp_ra_seq: bytes
        unitig: bytes

        for i in range(len(unitig_list)):
            RAlists_T.append([])         # for each unitig, there is another of: list RAs
            RAlists_C.append([])

        # assign RAs to unitigs

        for tmp_ra in self.RAlists[0]:
            flag = 0
            tmp_ra_seq = tmp_ra["SEQ"]
            for i in range(len(unitig_list)):
                unitig = unitig_list[i]
                if tmp_ra_seq in unitig:
                    flag = 1
                    RAlists_T[i].append(tmp_ra)
                    break
            if flag == 0:
                unmapped_RAlist_T.append(tmp_ra)
                # print "unmapped:", tmp_ra["SEQ"]

        for tmp_ra in self.RAlists[1]:
            flag = 0
            tmp_ra_seq = tmp_ra["SEQ"]
            for i in range(len(unitig_list)):
                unitig = unitig_list[i]
                if tmp_ra_seq in unitig:
                    flag = 1
                    RAlists_C[i].append(tmp_ra)
                    break
            if flag == 0:
                unmapped_RAlist_C.append(tmp_ra)
                # print "unmapped:", tmp_ra["SEQ"]

        # if unmapped_RAlist_T:
        # unmapped_ra_collection = RACollection(self.chrom, self.peak, unmapped_RAlist_T, unmapped_RAlist_C)
        return [RAlists_T, RAlists_C, unmapped_RAlist_T, unmapped_RAlist_C]

    @cython.cfunc
    def add_to_unitig_list(self, unitig_list, unitigs_2nd) -> list:
        """
        """
        i: cython.int
        j: cython.int
        flag: cython.int
        u0: bytes
        u1: bytes
        new_unitig_list: list

        new_unitig_list = []

        for i in range(len(unitigs_2nd)):
            # initial value: can't be found in unitig_list
            flag = 0
            u0 = unitigs_2nd[i]
            for j in range(len(unitig_list)):
                u1 = unitig_list[j]
                if u1.find(u0) != -1:
                    flag = 1
                    break
                u1 = u1[::-1].translate(__DNACOMPLEMENT__)
                if u1.find(u0) != -1:
                    flag = 1
                    break
            if not flag:
                new_unitig_list.append(u0)
        new_unitig_list.extend(unitig_list)
        return new_unitig_list

    @cython.ccall
    def build_unitig_collection(self, fermiMinOverlap):
        """unitig_and: list tuple_alns are in the same order!

        return UnitigCollection object.

        """
        start: cython.long
        end: cython.long
        unitigs_2nd: list
        # u: bytes
        # target_alns: list
        reference_alns: list
        aln_scores: list
        markup_alns: list
        # target_alns_2nd: list
        # reference_alns_2nd: list
        # aln_scores_2nd: list
        RAlists_T: list = []  # lists of of: list RAs of ChIP mapped to each unitig
        RAlists_C: list = []
        unmapped_RAlist_T: list = []
        unmapped_RAlist_C: list = []
        # tmp_unitig_seq: bytes
        tmp_reference_seq: bytes
        tmp_unitig_aln: bytes
        tmp_reference_aln: bytes

        i: cython.int
        j: cython.int
        left_padding_ref: cython.long
        right_padding_ref: cython.long
        left_padding_unitig: cython.long
        right_padding_unitig: cython.long
        ura_list: list = []
        unmapped_ra_collection: RACollection
        # flag: cython.int = 0
        # n_unmapped: cython.int
        n_unitigs_0: cython.int
        n_unitigs_1: cython.int

        # first round of assembly
        # print (" First round to assemble unitigs")
        unitig_list = self.fermi_assemble(fermiMinOverlap, opt_flag=0x80)
        if len(unitig_list) == 0:
            return 0

        n_unitigs_0 = -1
        n_unitigs_1 = len(unitig_list)
        # print " # of Unitigs:", n_unitigs_1
        # print " Map reads to unitigs"
        (unitig_alns, reference_alns, aln_scores, markup_alns) = self.align_unitig_to_REFSEQ(unitig_list)

        self.verify_alns(unitig_list,
                         unitig_alns,
                         reference_alns,
                         aln_scores,
                         markup_alns)
        if len(unitig_list) == 0:
            # if stop here, it raises a flag that the region may
            # contain too many mismapped reads, we return -1
            return -1
        # print (" # of Unitigs:", n_unitigs_1)

        # assign RAs to unitigs
        [RAlists_T, RAlists_C, unmapped_RAlist_T, unmapped_RAlist_C] = self.remap_RAs_w_unitigs(unitig_list)
        # prunmapped_ra_collection: cython.int.get_FASTQ().decode()

        # n_unmapped = len(unmapped_RAlist_T) + len(unmapped_RAlist_C)

        while len(unmapped_RAlist_T) > 0 and n_unitigs_1 != n_unitigs_0:
            # if there are unmapped reads AND we can get more unitigs
            # from last round of assembly, do assembly again

            # print (" # of RAs not mapped, will be assembled again:", n_unmapped)
            n_unitigs_0 = n_unitigs_1
            # another round of assembly
            unmapped_ra_collection = RACollection(self.chrom,
                                                  self.peak,
                                                  unmapped_RAlist_T,
                                                  unmapped_RAlist_C)
            unitigs_2nd = unmapped_ra_collection.fermi_assemble(fermiMinOverlap,
                                                                opt_flag=0x80)

            if unitigs_2nd:
                unitig_list = self.add_to_unitig_list(unitig_list, unitigs_2nd)
                n_unitigs_1 = len(unitig_list)
                # print " # of Unitigs:", n_unitigs_1
                # print " Map reads to unitigs"
                (unitig_alns, reference_alns, aln_scores, markup_alns) = self.align_unitig_to_REFSEQ(unitig_list)
                self.verify_alns(unitig_list,
                                 unitig_alns,
                                 reference_alns,
                                 aln_scores,
                                 markup_alns)
                [RAlists_T, RAlists_C, unmapped_RAlist_T, unmapped_RAlist_C] = self.remap_RAs_w_unitigs(unitig_list)
                # n_unmapped = len(unmapped_RAlist_T) + len(unmapped_RAlist_C)
            # else:
            #    for r in unmapped_RAlist_T:
            #        prr: cython.int.get_FASTQ().decode().lstrip()

            # print (" # of RAs not mapped, will be assembled again with 1/2 of fermiMinOverlap:", n_unmapped)
            # another round of assembly
            unmapped_ra_collection = RACollection(self.chrom,
                                                  self.peak,
                                                  unmapped_RAlist_T,
                                                  unmapped_RAlist_C)
            unitigs_2nd = unmapped_ra_collection.fermi_assemble(fermiMinOverlap/2,
                                                                opt_flag=0x80)

            if unitigs_2nd:
                unitig_list = self.add_to_unitig_list(unitig_list, unitigs_2nd)
                n_unitigs_1 = len(unitig_list)
                # print " # of Unitigs:", n_unitigs_1
                # print " Map reads to unitigs"
                (unitig_alns, reference_alns, aln_scores, markup_alns) = self.align_unitig_to_REFSEQ(unitig_list)
                self.verify_alns(unitig_list,
                                 unitig_alns,
                                 reference_alns,
                                 aln_scores,
                                 markup_alns)
                [RAlists_T, RAlists_C, unmapped_RAlist_T, unmapped_RAlist_C] = self.remap_RAs_w_unitigs(unitig_list)
                # n_unmapped = len(unmapped_RAlist_T) + len(unmapped_RAlist_C)
            # else:
            #    for r in unmapped_RAlist_T:
            #        prr: cython.int.get_FASTQ().decode().lstrip()
            if len(unitig_list) == 0:
                raise Exception("Shouldn't reach here")
            # print (" # of Unitigs:", n_unitigs_1)

        if len(unitig_list) == 0:
            return None
        # print (" Final round: # of Unitigs:", len(unitig_list))
        # print (" Final round: # of RAs not mapped:", n_unmapped)

        start = min(self.left, self.RAs_left)
        end = max(self.right, self.RAs_right)

        # create UnitigCollection
        for i in range(len(unitig_list)):
            #b'---------------------------AAATAATTTTATGTCCTTCAGTACAAAAAGCAGTTTCAACTAAAACCCAGTAACAAGCTAGCAATTCCTTTTAAATGGTGCTACTTCAAGCTGCAGCCAGGTAGCTTTTTATTACAAAAAATCCCACAGGCAGCCACTAGGTGGCAGTAACAGGCTTTTGCCAGCGGCTCCAGTCAGCATGGCTTGACTGTGTGCTGCAGAAACTTCTTAAATCGTCTGTGTTTGGGACTCGTGGGGCCCCACAGGGCTTTACAAGGGCTTTTTAATTTCCAAAAACATAAAACAAAAAAA--------------'
            #b'GATATAAATAGGATGTTATGAGTTTTCAAATAATTTTATGTCCTTCAGTACAAAAAGCAGTTTCAACTAAAACCCAGTAACAAGCTAGCAATTCCTTTTAAATGGTGCTACTTCAAGCTGCAGCCAGGTAGCTTTTTATTACAAAAA-TCCCACAGGCAGCCACTAGGTGGCAGTAACAGGCTTTTGCCAGCGGCTCCAGTCAGCATGGCTTGACTGTGTGCTGCAGAAACTTCTTAAATCGTCTGTGTTTGGGACTCGTGGGGCCCCACAGGGCTTTACAAGGGCTTTTTAATTTCCAAAAACATAAAACAAAAAAAAATACAAATGTATT'
            tmp_unitig_aln = unitig_alns[i]
            tmp_reference_aln = reference_alns[i]
            # tmp_unitig_seq = tmp_unitig_aln.replace(b'-',b'')
            tmp_reference_seq = tmp_reference_aln.replace(b'-', b'')

            # prtmp_unitig_aln: cython.int
            # prtmp_reference_aln: cython.int
            # prtmp_unitig_seq: cython.int
            # prtmp_reference_aln: cython.int

            # find the position on self.peak_refseq_ext
            left_padding_ref = self.peak_refseq_ext.find(tmp_reference_seq)  # this number of nts should be skipped on refseq_ext from left
            right_padding_ref = len(self.peak_refseq_ext) - left_padding_ref - len(tmp_reference_seq)  # this number of nts should be skipped on refseq_ext from right

            # now, decide the lpos and rpos on reference of this unitig
            # first, trim left padding '-'
            left_padding_unitig = len(tmp_unitig_aln) - len(tmp_unitig_aln.lstrip(b'-'))
            right_padding_unitig = len(tmp_unitig_aln) - len(tmp_unitig_aln.rstrip(b'-'))

            tmp_lpos = start + left_padding_ref
            tmp_rpos = end - right_padding_ref

            for j in range(left_padding_unitig):
                if tmp_reference_aln[j] != b'-':
                    tmp_lpos += 1
            for j in range(1, right_padding_unitig + 1):
                if tmp_reference_aln[-j] != b'-':
                    tmp_rpos -= 1

            tmp_unitig_aln = tmp_unitig_aln[left_padding_unitig:(len(tmp_unitig_aln)-right_padding_unitig)]
            tmp_reference_aln = tmp_reference_aln[left_padding_unitig:(len(tmp_reference_aln)-right_padding_unitig)]

            ura_list.append(UnitigRAs(self.chrom, tmp_lpos, tmp_rpos, tmp_unitig_aln, tmp_reference_aln, [RAlists_T[i], RAlists_C[i]]))

        return UnitigCollection(self.chrom, self.peak, ura_list)



#!/usr/bin/env python3
"""Description:

Setup script for MACS -- Model Based Analysis for ChIP-Seq data

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

import sys
import os
import re
from setuptools import setup, Extension
from Cython.Build import cythonize
import subprocess
import sysconfig
import numpy

# get MACS version
exec(open("MACS3/Utilities/Constants.py").read())


def main():
    if sys.version_info < (3, 9):
        sys.stderr.write("CRITICAL: Python version must >= 3.9!\n")
        sys.exit(1)

    # NumPy include dir
    numpy_include_dir = [numpy.get_include()]

    # CFLAG
    extra_c_args = ["-w", "-O3", "-g0"]
    # CFLAG for fermi-lite related codes
    clang = False
    icc = False
    new_gcc = False
    try:
        if os.environ['CC'] == "clang":
            clang = True
    except KeyError:
        pass

    if not clang:
        try:
            gcc_version_check = subprocess.check_output([os.environ.get('CC', 'gcc'), "--version"], universal_newlines=True)
            if gcc_version_check.find("clang") != -1:
                clang = True
            else:
                gcc_version_check = gcc_version_check.split('\n')[0]  # get the first line
                m = re.search(r"\s+(\d+\.\d+)\.\d+", gcc_version_check)
                if m:
                    gcc_version = float(m[1])
                    if gcc_version > 4.8:
                        new_gcc = True
        except subprocess.CalledProcessError:
            pass

    try:
        if os.environ['CC'] == "icc":
            icc = True
    except KeyError:
        pass

    extra_c_args_for_fermi = ["-std=gnu99", "-DUSE_SIMDE",
                              "-DSIMDE_ENABLE_NATIVE_ALIASES"]

    if icc or sysconfig.get_config_vars()['CC'] == 'icc':
        extra_c_args_for_fermi.extend(['-qopenmp-simd',
                                       '-DSIMDE_ENABLE_OPENMP'])
    elif new_gcc or clang or sysconfig.get_config_vars()['CC'] == 'clang':
        extra_c_args_for_fermi.extend(['-fopenmp-simd',
                                       '-DSIMDE_ENABLE_OPENMP'])

    # extensions, those have to be processed by Cython
    ext_modules = [Extension("MACS3.Signal.HMMR_EM",
                             ["MACS3/Signal/HMMR_EM.py"],
                             libraries=["m"],
                             include_dirs=numpy_include_dir,
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.HMMR_Signal_Processing",
                             ["MACS3/Signal/HMMR_Signal_Processing.py"],
                             libraries=["m"],
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.HMMR_HMM",
                             ["MACS3/Signal/HMMR_HMM.py"],
                             libraries=["m"],
                             include_dirs=numpy_include_dir,
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.Prob",
                             ["MACS3/Signal/Prob.py"],
                             libraries=["m"],
                             include_dirs=numpy_include_dir,
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.Region",
                             ["MACS3/Signal/Region.py"],
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.Pileup",
                             ["MACS3/Signal/Pileup.py",
                              "MACS3/Signal/cPosValCalculation.c"],
                             include_dirs=numpy_include_dir,
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.PileupV2",
                             ["MACS3/Signal/PileupV2.py"],
                             include_dirs=numpy_include_dir,
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.PeakModel",
                             ["MACS3/Signal/PeakModel.py"],
                             include_dirs=numpy_include_dir,
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.PeakDetect",
                             ["MACS3/Signal/PeakDetect.py"],
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.SignalProcessing",
                             ["MACS3/Signal/SignalProcessing.py"],
                             include_dirs=numpy_include_dir,
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.FixWidthTrack",
                             ["MACS3/Signal/FixWidthTrack.py"],
                             include_dirs=numpy_include_dir,
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.PairedEndTrack",
                             ["MACS3/Signal/PairedEndTrack.py"],
                             include_dirs=numpy_include_dir,
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.BedGraph",
                             ["MACS3/Signal/BedGraph.py"],
                             libraries=["m"],
                             include_dirs=numpy_include_dir,
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.ScoreTrack",
                             ["MACS3/Signal/ScoreTrack.py"],
                             libraries=["m"],
                             include_dirs=numpy_include_dir,
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.CallPeakUnit",
                             ["MACS3/Signal/CallPeakUnit.py"],
                             libraries=["m"],
                             include_dirs=numpy_include_dir,
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.VariantStat",
                             ["MACS3/Signal/VariantStat.py"],
                             include_dirs=numpy_include_dir,
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.ReadAlignment",
                             ["MACS3/Signal/ReadAlignment.py"],
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.RACollection",
                             ["MACS3/Signal/RACollection.py",
                              "MACS3/fermi-lite/bfc.c",
                              "MACS3/fermi-lite/bseq.c",
                              "MACS3/fermi-lite/bubble.c",
                              "MACS3/fermi-lite/htab.c",
                              "MACS3/fermi-lite/ksw.c",
                              "MACS3/fermi-lite/kthread.c",
                              "MACS3/fermi-lite/mag.c",
                              "MACS3/fermi-lite/misc.c",
                              "MACS3/fermi-lite/mrope.c",
                              "MACS3/fermi-lite/rld0.c",
                              "MACS3/fermi-lite/rle.c",
                              "MACS3/fermi-lite/rope.c",
                              "MACS3/fermi-lite/unitig.c",
                              "MACS3/Signal/swalign.c"],
                             libraries=["m", "z"],
                             include_dirs=["./",
                                           "./MACS3/fermi-lite/",
                                           "./MACS3/Signal/"],
                             extra_compile_args=extra_c_args+extra_c_args_for_fermi),
                   Extension("MACS3.Signal.UnitigRACollection",
                             ["MACS3/Signal/UnitigRACollection.py"],
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.PosReadsInfo",
                             ["MACS3/Signal/PosReadsInfo.py"],
                             include_dirs=numpy_include_dir,
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.Signal.PeakVariants",
                             ["MACS3/Signal/PeakVariants.py"],
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.IO.Parser",
                             ["MACS3/IO/Parser.py"],
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.IO.PeakIO",
                             ["MACS3/IO/PeakIO.py"],
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.IO.BedGraphIO",
                             ["MACS3/IO/BedGraphIO.py"],
                             extra_compile_args=extra_c_args),
                   Extension("MACS3.IO.BAM",
                             ["MACS3/IO/BAM.py",], libraries=["m"],
                             extra_compile_args=extra_c_args)]

    setup(version=MACS_VERSION,
          package_dir={'MACS3': 'MACS3'},
          packages=['MACS3', 'MACS3.IO', 'MACS3.Signal', 'MACS3.Commands', 'MACS3.Utilities'],
          package_data={'MACS3': ['*.pxd']},
          scripts=['bin/macs3', ],
          ext_modules=cythonize(ext_modules))


if __name__ == '__main__':
    main()



# cython: language_level=3
# cython: profile=True
# Time-stamp: <2025-04-11 13:26:28 Tao Liu>

"""Module Description:

New pileup algorithm based on p-v array idea. It's modified from
original algorithm in MACS2 proposed by Jie Wang. Now we allow
different weights for different genomic ranges, and simplify the
approach. The basic idea is to remember for each position, how we
modify the 'pileup' value while walking on the chromosome.

For a genomic range r_i covering genomic positions from `s_i` to `e_i`
to be piled up, we assign the start position a value `w_i`, and the
end -`w_i`, so we will use a tuple of position and weight to remember
these operations: (`s_i`, `w_i`) and (`e_i`, `-w_i`). Then all N
ranges will be made into an array (2D) of position and weights as:

PV = [ (s_0, w_0), (e_0, -w_0), (s_1, w_1), (e_1, -w_1), ... (s_i, w_i),
       (e_i, -w_i), ..., (s_N, w_N), (e_N, -w_N) ]

Then the array PV will be sorted by the first dimension, aka the
position, no matter the position is from start or end positions from
ranges.

PV_sorted = [ (p_0, v_0), (p_1, v_1), ... , (p_i, v_i), ..., (p_{2N}, v_{2N}) ]

The pileup algorithm to produce a bedGraph style pileup (another p-v
array as in Pileup.py) can be simply described as:

set the initial pileup z as 0 or a given value, and a start position s
as 0, and an end position e as not-defined.

for i from 0 to 2N in PV_sorted:
    1: z = z + v_i
    2: e = p_i
    3: save the pileup from position s to e is z,
       in bedGraph style is to only save (e, z)
    4: s = e

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""
# ------------------------------------
import numpy as np
import cython
import cython.cimports.numpy as cnp

# ------------------------------------
# from cython.cimports.libc.stdlib import malloc, free, qsort

# ------------------------------------
# utility internal functions
# ------------------------------------


@cython.ccall
def mapping_function_always_1(L: cython.int, R: cython.int) -> cython.float:
    # always return 1, useful while the weight is already 1, or in
    # case of simply piling up fragments for coverage.
    return 1.0


@cython.cfunc
def clean_up_ndarray(x: cnp.ndarray):
    """ Clean up numpy array in two steps
    """
    i: cython.long
    i = x.shape[0] // 2
    x.resize(100000 if i > 100000 else i, refcheck=False)
    x.resize(0, refcheck=False)
    return


@cython.cfunc
def make_PV_from_LR(LR_array: cnp.ndarray,
                    mapping_func=mapping_function_always_1) -> cnp.ndarray:
    """Make sorted PV array from a LR array for certain chromosome in a
    PETrackI object. The V/weight will be assigned as
    `mapping_func( L, R )` or simply 1 if mapping_func is the default.

    LR array is an np.ndarray as with dtype
    [('l','i4'),('r','i4')] with length of N

    PV array is an np.ndarray with
    dtype=[('p','i4'),('v','f4')] with length of 2N
    """
    l_LR: cython.ulong
    l_PV: cython.ulong
    i: cython.ulong
    L: cython.int
    R: cython.int
    PV: cnp.ndarray

    l_LR = LR_array.shape[0]
    l_PV = 2 * l_LR
    PV = np.zeros(shape=l_PV, dtype=[('p', 'i4'), ('v', 'f4')])
    for i in range(l_LR):
        (L, R) = LR_array[i]
        PV[i*2] = (L, mapping_func(L, R))
        PV[i*2 + 1] = (R, -1.0 * mapping_func(L, R))
    PV.sort(order='p')
    return PV


@cython.cfunc
def make_PV_from_LRC(LRC_array: cnp.ndarray,
                     mapping_func=mapping_function_always_1) -> cnp.ndarray:
    """Make sorted PV array from a LR array for certain chromosome in a
    PETrackII object. The V/weight will be assigned as
    `mapping_func( L, R )` or simply 1 if mapping_func is the default.

    LRC array is an np.ndarray as with dtype
    [('l','i4'),('r','i4'),('c','u2')] with length of N

    PV array is an np.ndarray with
    dtype=[('p','i4'),('v','f4')] with length of 2N
    """
    l_LRC: cython.ulong
    l_PV: cython.ulong
    i: cython.ulong
    L: cython.int
    R: cython.int
    C: cython.ushort
    PV: cnp.ndarray

    l_LRC = LRC_array.shape[0]
    l_PV = 2 * l_LRC
    PV = np.zeros(shape=l_PV, dtype=[('p', 'i4'), ('v', 'f4')])
    for i in range(l_LRC):
        (L, R, C) = LRC_array[i]
        PV[i*2] = (L, C*mapping_func(L, R))
        PV[i*2 + 1] = (R, -1.0 * C * mapping_func(L, R))
    PV.sort(order='p')
    return PV


@cython.cfunc
def make_PV_from_PN(P_array: cnp.ndarray, N_array: cnp.ndarray,
                    extsize: cython.int) -> cnp.ndarray:
    """Make sorted PV array from two arrays for certain chromosome in
    a FWTrack object. P_array is for the 5' end positions in plus
    strand, and N_array is for minus strand. We don't support weight
    in this case since all positions should be extended with a fixed
    'extsize'.

    P_array or N_array is an np.ndarray with dtype='i4'

    PV array is an np.ndarray with
    dtype=[('p','i4'),('v','f4')] with length of 2N
    """
    l_PN: cython.ulong
    l_PV: cython.ulong
    i: cython.ulong
    L: cython.int
    R: cython.int
    PV: cnp.ndarray

    l_PN = P_array.shape[0]
    assert l_PN == N_array.shape[0]
    l_PV = 4 * l_PN
    PV = np.zeros(shape=l_PV, dtype=[('p', 'i4'), ('v', 'f4')])
    for i in range(l_PN):
        L = P_array[i]
        R = L + extsize
        PV[i*2] = (L, 1)
        PV[i*2 + 1] = (R, -1)
    for i in range(l_PN):
        R = N_array[i]
        L = R - extsize
        PV[(l_PN + i)*2] = (L, 1)
        PV[(l_PN + i)*2 + 1] = (R, -1)
    PV.sort(order='p')
    return PV


@cython.cfunc
def pileup_PV(PV_array: cnp.ndarray) -> cnp.ndarray:
    """The pileup algorithm to produce a bedGraph style pileup (another
    p-v array as in Pileup.py) can be simply described as:

    set the initial pileup z as 0 or a given value, and a start
    position s as 0, and an end position e as not-defined.

    for i from 0 to 2N in PV_sorted:
        z = z + v_i
        e = p_i
        save the pileup from position s to e is z --  in bedGraph style is to only save (e, z)
        s = e
    """
    z: cython.float
    e: cython.int
    s: cython.int
    c: cython.int
    v: cython.float
    i: cython.ulong
    pre_z: cython.float

    # this is in bedGraph style as in Pileup.pyx, p is the end of a
    # region, and v is the pileup value.
    pileup_PV: cnp.ndarray
    z = 0
    pre_z = -10000
    s = 0
    pileup_PV = np.zeros(shape=PV_array.shape[0], dtype=[('p', 'i4'),
                                                         ('v', 'f4')])
    c = 0
    for i in range(PV_array.shape[0]):
        e = PV_array[i]['p']
        v = PV_array[i]['v']
        # make sure only to record the final value for the same position
        if e != s:
            # merge the p-v pair with the previous pair if the same v is found
            if z == pre_z:
                pileup_PV[c-1]['p'] = e
            else:
                pileup_PV[c] = (e, z)
                c += 1
                pre_z = z
        z += v
        s = e
    pileup_PV.resize(c, refcheck=False)
    # assert z == 0
    return pileup_PV

# ------------------------------------
# public python functions
# ------------------------------------


@cython.ccall
def pileup_from_LR_hmmratac(LR_array: cnp.ndarray,
                            mapping_dict: dict) -> cnp.ndarray:
    """Pile up LR fragments using HMMRATAC
    """
    # this function is specifically designed for piling up fragments
    # for `hmmratac`.
    #
    # As for `hmmratac`, the weight depends on the length of the
    # fragment, aka, value of R-L. Therefore, we need a mapping_dict
    # for mapping length to weight.
    l_LR: cython.ulong
    l_PV: cython.ulong
    i: cython.ulong
    L: cython.int
    R: cython.int
    PV: cnp.ndarray
    pileup: cnp.ndarray

    l_LR = LR_array.shape[0]
    l_PV = 2 * l_LR
    PV = np.zeros(shape=l_PV, dtype=[('p', 'i4'), ('v', 'f4')])
    for i in range(l_LR):
        (L, R) = LR_array[i]
        PV[i*2] = (L, mapping_dict[R - L])
        PV[i*2 + 1] = (R, -1 * mapping_dict[R - L])
    PV.sort(order='p')
    pileup = pileup_PV(PV)
    clean_up_ndarray(PV)
    return pileup


@cython.ccall
def pileup_from_LR(LR_array: cnp.ndarray,
                   mapping_func=mapping_function_always_1) -> cnp.ndarray:
    """This function will pile up the ndarray containing left and
    right positions, which is typically from PETrackI object. It's
    useful when generating the pileup of a single chromosome is
    needed.

    User needs to provide a numpy array of left and right positions,
    with dtype=[('l','i4'),('r','i4')]. User also needs to
    provide a mapping function to map the left and right position to
    certain weight.

    """
    PV_array: cnp.ndarray
    pileup: cnp.ndarray

    PV_array = make_PV_from_LR(LR_array, mapping_func=mapping_func)
    pileup = pileup_PV(PV_array)
    clean_up_ndarray(PV_array)
    return pileup


@cython.ccall
def pileup_from_LRC(LRC_array: cnp.ndarray,
                    mapping_func=mapping_function_always_1) -> cnp.ndarray:
    """This function will pile up the ndarray containing left and
    right positions and the counts, which is typically from PETrackII
    object. It's useful when generating the pileup of a single
    chromosome is needed.

    User needs to provide a numpy array of left and right positions
    and the counts, with
    dtype=[('l','i4'),('r','i4'),('c','u2')]. User also needs to
    provide a mapping function to map the left and right position to
    certain weight.

    """
    PV_array: cnp.ndarray
    pileup: cnp.ndarray

    PV_array = make_PV_from_LRC(LRC_array, mapping_func=mapping_func)
    pileup = pileup_PV(PV_array)
    clean_up_ndarray(PV_array)
    return pileup


@cython.ccall
def pileup_from_PN(P_array: cnp.ndarray, N_array: cnp.ndarray,
                   extsize: cython.int) -> cnp.ndarray:
    """This function will pile up the ndarray containing plus
    (positive) and minus (negative) positions of all reads, which is
    typically from FWTrackI object. It's useful when generating the
    pileup of a single chromosome is needed.

    """
    PV_array: cnp.ndarray
    pileup: cnp.ndarray

    PV_array = make_PV_from_PN(P_array, N_array, extsize)
    pileup = pileup_PV(PV_array)
    clean_up_ndarray(PV_array)
    return pileup



# Time-stamp: <2024-10-02 16:42:02 Tao Liu>

"""Description: combine replicates

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

import os

from MACS3.IO import BedGraphIO
from MACS3.Utilities.OptValidator import opt_validate_cmbreps

# ------------------------------------
# constants
# ------------------------------------

# ------------------------------------
# Misc functions
# ------------------------------------

# ------------------------------------
# Main function
# ------------------------------------


def run(options):
    options = opt_validate_cmbreps(options)
    info = options.info
    # warn = options.warn
    # debug = options.debug
    # error = options.error

    info("Read and build bedGraph for each replicate...")
    reps = []
    i = 1
    for ifile in options.ifile:
        info("Read file #%d" % i)
        reps.append(BedGraphIO.bedGraphIO(ifile).read_bedGraph())
        i += 1

    # first two reps
    info("combining tracks 1-%i with method '%s'" % (i - 1, options.method))
    cmbtrack = reps[0].overlie([reps[j] for j in range(1, i - 1)], func=options.method)

    # now output
    ofile = BedGraphIO.bedGraphIO(os.path.join(options.outdir, options.ofile), data=cmbtrack)
    info("Write bedGraph of combined scores...")
    ofile.write_bedGraph(name="%s_combined_scores" % (options.method.upper()), description="Scores calculated by %s" % (options.method.upper()))
    info("Finished '%s'! Please check '%s'!" % (options.method, ofile.bedGraph_filename))



# Time-stamp: <2024-10-02 16:16:49 Tao Liu>

"""Description: Naive call peaks from a single bedGraph track for
scores.

This code is free software; you can redistribute it and/or modify it
under the terms of the BSD License (see the file LICENSE included with
the distribution).
"""

# ------------------------------------
# python modules
# ------------------------------------
import os
from MACS3.IO import BedGraphIO
# ------------------------------------
# constants
# ------------------------------------

# ------------------------------------
# Misc functions
# ------------------------------------
from MACS3.Utilities.Logger import logging

logger = logging.getLogger(__name__)
debug = logger.debug
info = logger.info
error = logger.critical
warn = logger.warning
# ------------------------------------
# Classes
# ------------------------------------

# ------------------------------------
# Main function
# ------------------------------------


def run(options):
    info("Read and build bedGraph...")
    bio = BedGraphIO.bedGraphIO(options.ifile)
    btrack = bio.read_bedGraph(baseline_value=0)

    if options.cutoff_analysis:
        info("Analyze cutoff vs number of peaks/total length of peaks/average length of peak")
        cutoff_analysis_result = btrack.cutoff_analysis(int(options.maxgap), int(options.minlen), min_score=btrack.minvalue, max_score=int(options.cutoff_analysis_max), steps=int(options.cutoff_analysis_steps))
        info("Write report...")
        if options.ofile:
            fhd = open(os.path.join(options.outdir, options.ofile), 'w')
        else:
            fhd = open(os.path.join(options.outdir, "%s_l%d_g%d_cutoff_analysis.txt" % (options.oprefix,options.minlen, options.maxgap)), "w")
        fhd.write(cutoff_analysis_result)
        info("Done")
    else:
        info("Call peaks from bedGraph...")
        peaks = btrack.call_peaks(cutoff=float(options.cutoff), min_length=int(options.minlen), max_gap=int(options.maxgap), call_summits=options.call_summits)

        info("Write peaks...")
        if options.ofile:
            options.oprefix = options.ofile
            nf = open(os.path.join(options.outdir, options.ofile), 'w')
        else:
            nf = open(os.path.join(options.outdir, "%s_c%.1f_l%d_g%d_peaks.narrowPeak" % (options.oprefix, options.cutoff, options.minlen, options.maxgap)), "w")
        peaks.write_to_narrowPeak(nf, name=options.oprefix.encode(), name_prefix=(options.oprefix+"_narrowPeak").encode(), score_column="score", trackline=options.trackline)
        info("Done")
